import math
from typing import Tuple, Union

import numpy

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.filters.fft_convolve import fft_convolve
from dexp.processing.filters.kernels.gaussian import gaussian_kernel_nd
from dexp.processing.filters.kernels.wiener_butterworth import wiener_butterworth_kernel
from dexp.processing.utils.nan_to_zero import nan_to_zero
from dexp.processing.utils.normalise import normalise_functions


def lucy_richardson_deconvolution(image,
                                  psf,
                                  num_iterations: int = None,
                                  max_correction: float = None,
                                  power: float = 1.0,
                                  back_projection='tpsf',
                                  wb_cutoffs: Union[float, Tuple[float, ...], None] = 0.9,
                                  wb_beta: float = 0.05,
                                  wb_order: int = 2,
                                  padding: int = 0,
                                  padding_mode: str = 'edge',
                                  normalise_input: bool = True,
                                  normalise_minmax: Tuple[float, float] = None,
                                  clip_output: bool = False,
                                  blind_spot: int = 0,
                                  blind_spot_mode: str = 'median+uniform',
                                  blind_spot_axis_exclusion: Union[str, Tuple[int, ...]] = None,
                                  convolve_method=fft_convolve,
                                  internal_dtype=None):
    """
    Deconvolves an nD image given a point-spread-function.

    Parameters
    ----------
    image : image to deconvolve
    psf : point-spread-function (must have the same number of dimensions as image!)
    num_iterations : number of iterations
    max_correction : Lucy-Richardson correction will remain clamped within [1/mc, mc] (before back projection)
    power : power to elevate coorection (before back projection)
    back_projection : back projection operator to use: 'tpsf' or 'wb'.
    padding : padding (see numpy/cupy pad function)
    padding_mode : padding mode (see numpy/cupy pad function)
    normalise_input : This deconvolution code assumes values within [0, 1], by default input images are normalised to that range, but if already normalised, then normalisation can be ommited.
    normalise_minmax : Use the given tuple (min, max) for normalisation
    blind_spot : If zero, blind-spot is disabled. If blind_spot>0 it is active and the integer represents the blind-spot kernel support. A value of 3 or 5 are good and help reduce the impact of noise on deconvolution.
    blind_spot_mode : blind-spot mode, can be 'mean' or 'median'
    blind_spot_axis_exclusion : if None no axis is excluded from the blind-spot support kernel, otherwise if a tuple of ints is provided, these refer to the
    then the support kernel is clipped along these axis. For example for a 3D stack where the sampling along z (first axis) is poor, use: (0,) so that blind-spot kernel does not extend in z.
    convolve_method : convolution method to use
    clip_output : Clip output to input range, or not
    internal_dtype : dtype to use internally for computation.

    Returns
    -------
    Deconvolved image

    """
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    if image.ndim != psf.ndim:
        raise ValueError("The image and PSF must have same number of dimensions!")

    if internal_dtype is None:
        internal_dtype = numpy.float32

    if type(Backend.current()) is NumpyBackend:
        internal_dtype = numpy.float32

    original_dtype = image.dtype
    image = Backend.to_backend(image, dtype=internal_dtype)
    psf = Backend.to_backend(psf, dtype=internal_dtype)

    if blind_spot > 0:
        if 2 * (blind_spot // 2) == blind_spot:
            raise ValueError(f"Blind spot size must be an odd integer, blind_spot={blind_spot} is not!")

        if 'gaussian' in blind_spot_mode:
            full_kernel = gaussian_kernel_nd(ndim=image.ndim, size=blind_spot, sigma=max(1, blind_spot // 2))
        else:
            full_kernel = xp.ones(shape=(blind_spot,) * image.ndim, dtype=internal_dtype)

        if blind_spot_axis_exclusion is not None:
            c = blind_spot // 2

            slicing_n = [slice(None, None, None), ] * image.ndim
            slicing_p = [slice(None, None, None), ] * image.ndim
            for axis in blind_spot_axis_exclusion:
                slicing_n[axis] = slice(0, c, None)
                slicing_p[axis] = slice(c + 1, blind_spot, None)

            full_kernel[slicing_n] = 0
            full_kernel[slicing_p] = 0

        full_kernel /= full_kernel.sum()
        donut_kernel = full_kernel.copy()
        donut_kernel[(slice(blind_spot // 2, blind_spot // 2 + 1, None),) * image.ndim] = 0
        donut_kernel /= donut_kernel.sum()
        # psf_original = psf.copy()
        psf = sp.ndimage.convolve(psf, donut_kernel)
        psf /= psf.sum()

        if 'median' in blind_spot_mode:
            image = sp.ndimage.filters.median_filter(image, footprint=full_kernel)
        elif 'mean' in blind_spot_mode:
            image = sp.ndimage.convolve(image, donut_kernel)

        # from napari import Viewer, gui_qt
        # with gui_qt():
        #     def _c(array):
        #         return backend.to_numpy(array)
        #
        #     viewer = Viewer()
        #     #viewer.add_image(_c(image), name='image')
        #     viewer.add_image(_c(psf_original), name='psf_original')
        #     viewer.add_image(_c(donut_kernel), name='donut_kernel', color=False)
        #     viewer.add_image(_c(psf), name='psf')
        #     viewer.add_image(_c(sp.ndimage.convolve(psf, full_kernel)), name='psf_for_backproj')
        #     viewer.add_image(_c(back_projector), name='psf')

    if num_iterations is None:
        if back_projection == 'tpsf':
            num_iterations = 20
        elif back_projection == 'wb':
            num_iterations = 3

    if back_projection == 'tpsf':
        back_projector = xp.flip(psf)
    elif back_projection == 'wb':
        back_projector = wiener_butterworth_kernel(kernel=xp.flip(psf),
                                                   cutoffs=wb_cutoffs,
                                                   beta=wb_beta,
                                                   order=wb_order)
    else:
        raise ValueError(f"back projection mode: {back_projection} not supported.")

    # from napari import Viewer, gui_qt
    # with gui_qt():
    #     def _c(array):
    #         return backend.to_numpy(array)
    #     psf_f = xp.log1p(xp.absolute(xp.fft.fftshift(xp.fft.fftn(psf))))
    #     back_projector_f = xp.log1p(xp.absolute(xp.fft.fftshift(xp.fft.fftn(back_projector))))
    #     viewer = Viewer()
    #     viewer.add_image(_c(psf), name='psf')
    #     viewer.add_image(_c(back_projector), name='back_projector')
    #     viewer.add_image(_c(psf_f), name='psf_f', colormap='viridis')
    #     viewer.add_image(_c(back_projector_f), name='back_projector_f', colormap='viridis')

    norm_fun, denorm_fun = normalise_functions(image,
                                               minmax=normalise_minmax,
                                               do_normalise=normalise_input,
                                               clip=False,
                                               dtype=internal_dtype)

    image = norm_fun(image)

    if padding > 0:
        image = numpy.pad(image, pad_width=padding, mode=padding_mode)

    result = xp.full(
        image.shape,
        float(xp.mean(image)),
        dtype=internal_dtype
    )

    for i in range(num_iterations):
        # print(f"LR iteration: {i}")

        convolved = convolve_method(
            result,
            psf,
            mode='wrap',
        )

        relative_blur = image / convolved

        # replace Nans with zeros:
        # zeros = convolved == 0
        # relative_blur[zeros] = 0
        relative_blur = nan_to_zero(relative_blur, copy=False)

        if max_correction is not None:
            relative_blur[
                relative_blur > max_correction
                ] = max_correction
            relative_blur[relative_blur < 1 / max_correction] = (
                    1 / max_correction
            )

        multiplicative_correction = convolve_method(
            relative_blur,
            back_projector,
            mode='wrap',
        )

        if back_projection == 'wb':
            multiplicative_correction = xp.clip(multiplicative_correction,
                                                a_min=0, a_max=None,
                                                out=multiplicative_correction)

        if power != 1.0:
            multiplicative_correction = xp.clip(multiplicative_correction, 0, None, out=multiplicative_correction)
            multiplicative_correction **= (1 + (power - 1) / (math.sqrt(1 + i)))

        result *= multiplicative_correction

    del multiplicative_correction, relative_blur, back_projector, convolved, psf

    if clip_output:
        if normalise_input:
            result = xp.clip(result, 0, 1, out=result)
        else:
            result = xp.clip(result, xp.min(image), xp.max(image), out=result)

    result = denorm_fun(result)

    if padding > 0:
        slicing = (slice(padding, -padding),) * result.ndim
        result = result[slicing]

    result = result.astype(original_dtype, copy=False)

    # from napari import Viewer
    # import napari
    # with napari.gui_qt():
    #     def _c(array):
    #         return backend.to_numpy(array)
    #
    #     viewer = Viewer()
    #     viewer.add_image(_c(image1), name='image_1')
    #     viewer.add_image(_c(image1), name='image_2')

    return result



import os
from os import listdir
from os.path import exists, isfile, join, isdir
from typing import Tuple, Sequence, Union

import imageio
from arbol.arbol import asection, aprint
from joblib import Parallel, delayed

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.best_backend import BestBackend
from dexp.processing.color.blend import blend_color_images
from dexp.processing.color.scale_bar import insert_scale_bar
from dexp.processing.color.time_stamp import insert_time_stamp


def add_overlays_image_sequence(input_path: str,
                                output_path: str = None,
                                scale_bar: bool = True,
                                scale_bar_length_in_unit: float = 1,
                                scale_bar_pixel_scale: float = 1,
                                scale_bar_bar_height: int = 4,
                                scale_bar_translation: Union[str, Sequence[Tuple[Union[int, float], ...]]] = 'bottom_right',
                                scale_bar_unit: str = 'μm',
                                time_stamp: bool = True,
                                time_stamp_start_time: float = 0,
                                time_stamp_time_interval: float = 1,
                                time_stamp_translation: Union[str, Sequence[Tuple[Union[int, float], ...]]] = 'top_right',
                                time_stamp_unit: str = 's',
                                margin: float = 1,
                                color: Tuple[float, float, float, float] = None,
                                number_format: str = '{:.1f}',
                                font_name: str = "Helvetica",
                                font_size: float = 32,
                                mode: str = 'max',
                                overwrite: bool = False,
                                workers: int = -1,
                                workersbackend: str = 'threading',
                                device: int = 0):
    """
    Blends several RGB(A) image sequences together

    Parameters
    ----------
    input_path : Path to folder containing images in some (lexicographic) order.
    output_path : Path to save the blended images.
    scale_bar: True to insert scale bar.
    scale_bar_length_in_unit: Length of scale bar in the provided unit.
    scale_bar_pixel_scale: conversion factor from pixels to units -- what is the side length of a pixel/voxel in units.
    scale_bar_bar_height: Height of th scale bar in pixels
    scale_bar_translation: Positions of the scale bar in pixels in numpy order: (y, x). Can also be a string: 'bottom_left', 'bottom_right', 'top_left', 'top_right'.
    scale_bar_unit: Scale bar unit name.
    time_stamp: True to insert time stamp.
    time_stamp_start_time: Start time for time stamp
    time_stamp_time_interval: Time interval inn units of time between consecutive images.
    time_stamp_translation: Positions of the time stamp in pixels in numpy order: (y, x). Can also be a string: 'bottom_left', 'bottom_right', 'top_left', 'top_right'.
    time_stamp_unit: Time stamp time unit name.
    margin: margin around bar expressed in units relative to the text height
    color: Color of the bar and text as tuple of 4 values: (R, G, B, A)
    number_format: Format string to represent the start and end values.
    font_name: Font name.
    font_size: Font size in pixels.
    mode: Blending mode. See function 'blend_color_images' for available blending modes.
    overwrite : If True the output files are overwritten
    workers : Number of worker threads to spawn, if -1 then num workers = num devices', show_default=True)
    workersbackend : What backend to spawn workers with, can be ‘loky’ (multi-process) or ‘threading’ (multi-thread)
    device: Device on  which to run the overlay computation (if a non-CPU device is available).
    """

    # ensure folder exists:
    os.makedirs(output_path, exist_ok=True)

    # collect image files:
    if isdir(input_path):
        # path is folder:
        png_file_paths = [join(input_path, f) for f in listdir(input_path) if isfile(join(input_path, f)) and f.endswith('.png')]
        png_file_paths.sort()
    else:
        raise ValueError("Input path must be folder containing at least one image")

    if workers == -1:
        workers = min(8, os.cpu_count() // 2)

    # Number of timepoints:
    nb_timepoints = len(png_file_paths)

    # Load sample image:
    sample_image = imageio.imread(png_file_paths[0])

    if scale_bar:
        with asection("Applying scale bar..."):
            # First generate the scale bar itself:
            _, scale_bar_image = insert_scale_bar(sample_image,
                                                  length_in_unit=scale_bar_length_in_unit,
                                                  pixel_scale=scale_bar_pixel_scale,
                                                  bar_height=scale_bar_bar_height,
                                                  margin=margin,
                                                  translation=scale_bar_translation,
                                                  color=color,
                                                  number_format=number_format,
                                                  font_name=font_name,
                                                  font_size=font_size,
                                                  unit=scale_bar_unit,
                                                  mode=mode
                                                  )

    def _process(tp: int):
        with asection(f'Processing time point: {tp}'):

            # Output file:
            filename = f"frame_{tp:05}.png"
            filepath = join(output_path, filename)

            # Write file:
            if overwrite or not exists(filepath):
                with BestBackend(device, exclusive=True, enable_unified_memory=True):

                    # Get image path:
                    image_path = png_file_paths[tp]

                    # get image:
                    image = imageio.imread(image_path)

                    # Apply time stamp:
                    if time_stamp:
                        with asection("Applying time stamp..."):
                            image = insert_time_stamp(image=image,
                                                      time_point_index=tp,
                                                      nb_time_points=nb_timepoints,
                                                      start_time=time_stamp_start_time,
                                                      time_interval=time_stamp_time_interval,
                                                      unit=time_stamp_unit,
                                                      margin=margin,
                                                      translation=time_stamp_translation,
                                                      color=color,
                                                      number_format=number_format,
                                                      font_name=font_name,
                                                      font_size=font_size,
                                                      mode=mode)

                    # Apply scale bar:
                    image_with_scale_bar = blend_color_images(images=(image,
                                                                      scale_bar_image),
                                                              alphas=(1, 1),
                                                              modes=('max', mode))


                    with asection(f"Writing file: {filename} in folder: {output_path}"):
                        imageio.imwrite(filepath,
                                        Backend.to_numpy(image_with_scale_bar),
                                        compress_level=1)
            else:
                aprint(f"File: {filepath} already exists! use -w option to force overwrite...")

    with asection(f"Adding time-stamp ({insert_time_stamp}) and scale-bar ({insert_scale_bar}) to: {input_path}, and saving to {output_path}, for a total of {nb_timepoints} time points"):
        Parallel(n_jobs=workers, backend=workersbackend)(delayed(_process)(tp) for tp in range(nb_timepoints))
        aprint(f"Done!")



from typing import Sequence, Tuple

from arbol.arbol import aprint
from arbol.arbol import asection
from joblib import Parallel, delayed
from skimage.transform import downscale_local_mean

from dexp.datasets.base_dataset import BaseDataset
from dexp.optics.psf.standard_psfs import nikon16x08na, olympus20x10na
from dexp.processing.backends.backend import Backend
from dexp.processing.backends.best_backend import BestBackend
from dexp.processing.deconvolution.lr_deconvolution import lucy_richardson_deconvolution
from dexp.processing.utils.scatter_gather_i2i import scatter_gather_i2i


def dataset_deconv(dataset: BaseDataset,
                   path: str,
                   channels: Sequence[str],
                   slicing,
                   store: str,
                   compression: str,
                   compression_level: int,
                   overwrite: bool,
                   chunksize: Tuple[int],
                   method: str,
                   num_iterations: int,
                   max_correction: int,
                   power: float,
                   blind_spot: int,
                   back_projection: str,
                   objective: str,
                   numerical_aperture: float,
                   dxy: float,
                   dz: float,
                   xy_size: int,
                   z_size: int,
                   show_psf: bool,
                   scaling: Tuple[float],
                   workers: int,
                   workersbackend: str,
                   devices: Sequence[int],
                   check: bool,
                   stop_at_exception: bool = True,):
    from dexp.datasets.zarr_dataset import ZDataset
    mode = 'w' + ('' if overwrite else '-')
    dest_dataset = ZDataset(path, mode, store)

    sz, sy, sx = scaling
    aprint(f"Input images will be scaled by: (sz,sy,sx)={scaling}")

    for channel in dataset._selected_channels(channels):

        array = dataset.get_array(channel, per_z_slice=False, wrap_with_dask=True)

        if slicing is not None:
            array = array[slicing]

        shape = tuple(int(round(u*v)) for u, v in zip(array.shape, (1,)+scaling))
        chunks = ZDataset._default_chunks
        nb_timepoints = shape[0]

        dest_array = dest_dataset.add_channel(name=channel,
                                              shape=shape,
                                              dtype=array.dtype,
                                              chunks=chunks,
                                              codec=compression,
                                              clevel=compression_level)

        #This is not ideal but difficult to avoid right now:
        sxy = (sx+sy)/2

        psf_kwargs = {'dxy': dxy / sxy,
                      'dz': dz / sz,
                      'xy_size': int(round(xy_size * sxy)),
                      'z_size': int(round(z_size * sz))}

        aprint(f"psf_kwargs: {psf_kwargs}")

        if numerical_aperture is not None:
            aprint(f"Numerical aperture overridden to a value of: {numerical_aperture}")
            psf_kwargs['NA'] = numerical_aperture

        if objective == 'nikon16x08na':
            psf_kernel = nikon16x08na(**psf_kwargs)
        elif objective == 'olympus20x10na':
            psf_kernel = olympus20x10na(**psf_kwargs)

        if show_psf:
            from napari import gui_qt, Viewer
            with gui_qt():
                viewer = Viewer(title=f"DEXP | viewing PSF with napari", ndisplay=3)
                viewer.add_image(psf_kernel)


        def process(tp, device):

            try:
                with asection(f"Loading channel: {channel} for time point {tp}/{nb_timepoints}"):
                    tp_array = array[tp].compute()

                with BestBackend(device, exclusive=True):

                    if sz != 1.0 or sy != 1.0 or sx != 1.0:
                        with asection(f"Applying scaling {(sz, sy, sx)} to image."):
                            sp = Backend.get_sp_module()
                            tp_array = Backend.to_backend(tp_array)
                            tp_array = sp.ndimage.interpolation.zoom(tp_array, zoom=(sz, sy, sx), order=1)
                            tp_array = Backend.to_numpy(tp_array)

                    if method == 'lr':
                        min_value = tp_array.min()
                        max_value = tp_array.max()

                        def f(image):

                            return lucy_richardson_deconvolution(image=image,
                                                                 psf=psf_kernel,
                                                                 num_iterations=num_iterations,
                                                                 max_correction=max_correction,
                                                                 normalise_minmax=(min_value, max_value),
                                                                 power=power,
                                                                 blind_spot=blind_spot,
                                                                 blind_spot_mode='median+uniform',
                                                                 blind_spot_axis_exclusion=(0,),
                                                                 back_projection=back_projection
                                                                 )

                        margins = max(xy_size, z_size)
                        with asection(f"LR Deconvolution of image of shape: {tp_array.shape}, with chunk size: {chunksize}, margins: {margins} "):
                            aprint(f"Number of iterations: {num_iterations}, back_projection:{back_projection}, ")
                            tp_array = scatter_gather_i2i(f, tp_array, chunks=chunksize, margins=margins)
                    else:
                        raise ValueError(f"Unknown deconvolution mode: {method}")

                    with asection(f"Moving array from backend to numpy."):
                        tp_array = Backend.to_numpy(tp_array, dtype=dest_array.dtype, force_copy=False)

                with asection(f"Saving deconvolved stack for time point {tp}, shape:{tp_array.shape}, dtype:{array.dtype}"):
                    dest_dataset.write_stack(channel=channel,
                                             time_point=tp,
                                             stack_array=tp_array)

                aprint(f"Done processing time point: {tp}/{nb_timepoints} .")

            except Exception as error:
                aprint(error)
                aprint(f"Error occurred while processing time point {tp} !")
                import traceback
                traceback.print_exc()

                if stop_at_exception:
                    raise error

        if workers == -1:
            workers = len(devices)
        aprint(f"Number of workers: {workers}")

        if workers > 1:
            Parallel(n_jobs=workers, backend=workersbackend)(delayed(process)(tp, devices[tp % len(devices)]) for tp in range(0, nb_timepoints))
        else:
            for tp in range(0, nb_timepoints):
                process(tp, devices[0])


    aprint(dest_dataset.info())
    if check:
        dest_dataset.check_integrity()

    # close destination dataset:
    dest_dataset.close()



from typing import Sequence, Union, List

import dask
from arbol import aprint
from dask.array import reshape

from dexp.datasets.base_dataset import BaseDataset


def dataset_view(input_dataset: BaseDataset,
                 channels: Sequence[str],
                 slicing,
                 aspect: float,
                 contrast_limits: Union[List[int], List[float]],
                 colormap: str,
                 name: str,
                 windowsize: int,
                 projections_only: bool,
                 volume_only: bool):
    # Annoying napari induced warnings:
    import warnings
    warnings.filterwarnings("ignore")

    from napari import gui_qt, Viewer
    from napari._qt.qthreading import thread_worker
    with gui_qt():
        viewer = Viewer(title=f"DEXP | viewing with napari: {name} ", ndisplay=2)
        viewer.grid.enabled = True
        viewer.window.resize(windowsize + 256, windowsize)

        for channel in channels:
            aprint(f"Channel '{channel}' shape: {input_dataset.shape(channel)}")
            aprint(input_dataset.info(channel))

            array = input_dataset.get_array(channel, wrap_with_dask=True)

            try:
                if not volume_only:
                    for axis in range(array.ndim - 1):
                        proj_array = input_dataset.get_projection_array(channel, axis=axis, wrap_with_dask=True)

                        shape = (proj_array.shape[0], 1,) + proj_array.shape[1:]
                        proj_array = reshape(proj_array, shape=shape)

                        if proj_array is not None:
                            proj_layer = viewer.add_image(proj_array,
                                                          name=channel + '_proj_' + str(axis),
                                                          contrast_limits=contrast_limits,
                                                          blending='additive',
                                                          colormap=colormap, )

                            if aspect is not None:
                                if axis == 0:
                                    proj_layer.scale = (1, 1)
                                elif axis == 1:
                                    proj_layer.scale = (aspect, 1)
                                elif axis == 2:
                                    proj_layer.scale = (aspect, 1)

                                aprint(f"Setting aspect ratio for projection (layer.scale={proj_layer.scale})")

            except KeyError:
                aprint("Warning: can't find projections!")

            if not projections_only:

                if slicing:
                    array = array[slicing]

                aprint(f"Adding array of shape={array.shape} and dtype={array.dtype} for channel '{channel}'.")

                # flip x for second camera:
                if 'C1' in channel:
                    array = dask.array.flip(array, -1)

                layer = viewer.add_image(array,
                                         name=channel,
                                         contrast_limits=contrast_limits,
                                         blending='additive',
                                         colormap=colormap,
                                         attenuation=0.04,
                                         rendering='attenuated_mip')

                if aspect is not None:
                    if array.ndim == 3:
                        layer.scale = (aspect, 1, 1)
                    elif array.ndim == 4:
                        layer.scale = (1, aspect, 1, 1)
                    aprint(f"Setting aspect ratio to {aspect} (layer.scale={layer.scale})")

                # For some reason some parameters refuse to be set, this solves it:
                @thread_worker
                def workaround_for_recalcitrant_parameters():
                    aprint("Setting 3D rendering parameters")
                    layer.attenuation = 0.02
                    layer.rendering = 'attenuated_mip'

                worker = workaround_for_recalcitrant_parameters()
                worker.start()



import os
from os.path import join
from typing import Sequence

from arbol.arbol import aprint, asection
from joblib import Parallel, delayed
from tifffile import memmap

from dexp.datasets.base_dataset import BaseDataset
from dexp.io.io import tiff_save


def dataset_tiff(dataset: BaseDataset,
                 output_path: str,
                 channels: Sequence[str],
                 slicing,
                 overwrite: bool,
                 project: bool,
                 one_file_per_first_dim: bool,
                 clevel: int,
                 workers: int,
                 workersbackend: str):
    selected_channels = dataset._selected_channels(channels)

    aprint(f"getting Dask arrays for channels {selected_channels}")
    arrays = list([dataset.get_array(channel, per_z_slice=False, wrap_with_dask=True) for channel in selected_channels])

    if slicing is not None:
        aprint(f"Slicing with: {slicing}")
        arrays = list([array[slicing] for array in arrays])
        aprint(f"Done slicing.")

    if project:
        # project is the axis for projection, but here we are not considering the T dimension anymore...
        aprint(f"Projecting along axis {project}")
        arrays = list([array.max(axis=project) for array in arrays])

    if workers == -1:
        workers = max(1, os.cpu_count() // abs(workers))
    aprint(f"Number of workers: {workers}")

    if one_file_per_first_dim:
        aprint(f"Saving one TIFF file for each tp (or Z if already sliced) to: {output_path}.")

        os.makedirs(output_path, exist_ok=True)

        def process(tp):
            with asection(f'Saving time point {tp}: '):
                for channel, array in zip(selected_channels, arrays):
                    tiff_file_path = join(output_path, f"file{tp}_{channel}.tiff")
                    if overwrite or not os.path.exists(tiff_file_path):
                        stack = array[tp].compute()
                        print(f"Writing time point: {tp} of shape: {stack.shape}, dtype:{stack.dtype} as TIFF file: '{tiff_file_path}', with compression: {clevel}")
                        tiff_save(tiff_file_path, stack, compress=clevel)
                        print(f"Done writing time point: {tp} !")
                    else:
                        print(f"File for time point (or z slice): {tp} already exists.")

        if workers > 1:
            Parallel(n_jobs=workers, backend=workersbackend)(delayed(process)(tp) for tp in range(0, arrays[0].shape[0]))
        else:
            for tp in range(0, arrays[0].shape[0]):
                process(tp)

    else:

        for channel, array in zip(selected_channels, arrays):
            if len(selected_channels) > 1:
                tiff_file_path = f"{output_path}_{channel}.tiff"
            else:
                tiff_file_path = f"{output_path}.tiff"

            if not overwrite and os.path.exists(tiff_file_path):
                aprint(f"File {tiff_file_path} already exists! Set option -w to overwrite.")
                return

            with asection(f"Saving array ({array.shape}, {array.dtype}) for channel {channel} into TIFF file at: {tiff_file_path}:"):
                memmap_image = memmap(tiff_file_path, shape=array.shape, dtype=array.dtype, bigtiff=True, imagej=True)

                def process(tp):
                    aprint(f"Processing time point {tp}")
                    stack = array[tp].compute()
                    memmap_image[tp] = stack

                if workers > 1:
                    Parallel(n_jobs=workers, backend=workersbackend)(delayed(process)(tp) for tp in range(0, array.shape[0]))
                else:
                    for tp in range(0, array.shape[0]):
                        process(tp)

                memmap_image.flush()
                del memmap_image

## NOTES: color coded max projection:
# > data = numpy.random.randint(0, 255, (256, 256, 3), 'uint8')
# >>> imwrite('temp.tif', data, photometric='color')
#
# https://colorcet.holoviz.org/user_guide/index.html
# https://github.com/MMesch/cmap_builder
#



from typing import Tuple, Sequence, Any, List

import numpy
from arbol.arbol import aprint
from dask.array import concatenate

from dexp.datasets.base_dataset import BaseDataset


# Configure multithreading for Dask:


class JoinedDataset(BaseDataset):

    def __init__(self, datasets: Sequence[BaseDataset]):
        """Instanciates a joined dataset.

        Parameters
        ----------
        datasets : sequence of datasets to join into one temporally concatenated dataset.

        Returns
        -------
        Joined dataset


        """

        super().__init__(dask_backed=False)

        self._dataset_list: List[BaseDataset] = list(datasets)

        aprint(f"dataset list: {self._dataset_list}")

        # First we make sure that the list is not empty:
        if len(self._dataset_list) == 0:
            raise ValueError("Dataset list is empty!")

        # Second we check if the same channels are present in all datasets:
        _dataset_zero = self._dataset_list[0]
        for i, dataset in enumerate(self._dataset_list):
            if set(dataset.channels()) != set(_dataset_zero.channels()):
                aprint(f"dataset #{i} has channels '{dataset.channels()}' but datatset #0 has channels: '{_dataset_zero.channels()}'")
                raise ValueError("All datasets must have the same exact channels!")

            # Third, we also check per channel if the shape and dtypes are the same
            for channel in _dataset_zero.channels():
                if _dataset_zero.shape(channel)[1:] != dataset.shape(channel)[1:]:
                    aprint(f"dataset #{i} channel {channel} has shape '{dataset.shape(channel)[1:]}' but datatset #0 channel {channel} has shape: '{_dataset_zero.shape(channel)[1:]}'")
                    raise ValueError("All datasets must have the same exact shape for the same channels! (except for time dimension!)")
                if _dataset_zero.dtype(channel) != dataset.dtype(channel):
                    aprint(f"dataset #{i} channel {channel} has dtype '{dataset.dtype(channel)}' but datatset #0 channel {channel} has dtype: '{_dataset_zero.dtype(channel)}'")
                    raise ValueError("All datasets must have the same exact dtype for the same channels!")

        # Build concatenated dask arrays:
        self._arrays = {}
        self._projection_arrays = {}
        for channel in self.channels():
            self._arrays[channel] = concatenate(dataset.get_array(channel, per_z_slice=False, wrap_with_dask=True) for dataset in self._dataset_list)

            for axis in range(len(self.shape(channel)) - 1):
                try:
                    self._projection_arrays[f'{channel}/{axis}'] = concatenate(dataset.get_projection_array(channel, axis=axis, wrap_with_dask=True) for dataset in self._dataset_list)
                except KeyError:
                    self._projection_arrays[f'{channel}/{axis}'] = None

    def close(self):
        for dataset in self._dataset_list:
            dataset.close()

    def check_integrity(self, channels: Sequence[str] = None) -> bool:
        for dataset in self._dataset_list:
            if not dataset.check_integrity(channels):
                return False
        return True

    def channels(self) -> Sequence[str]:
        return self._dataset_list[0].channels()

    def shape(self, channel: str) -> Sequence[int]:
        return (self.nb_timepoints(channel),) + tuple(self._dataset_list[0].shape(channel)[1:])

    def nb_timepoints(self, channel: str) -> int:
        return sum(dataset.nb_timepoints(channel) for dataset in self._dataset_list)

    def dtype(self, channel: str):
        return self._dataset_list[0].dtype(channel)

    def info(self, channel: str = None) -> str:
        if channel:
            info_str = f"Channel: '{channel}', nb time points: {self.shape(channel)[0]}, shape: {self.shape(channel)[1:]}, joined from {len(self._dataset_list)} datasets."
            info_str += "\n"
            return info_str
        else:
            info_str = f"Joined dataset of length: {len(self._dataset_list)} \n"
            info_str += "\n\n"
            info_str += "Channels: \n"
            for channel in self.channels():
                info_str += "  └──" + self.info(channel) + "\n\n"
                for i, dataset in enumerate(self._dataset_list):
                    info_str += f"      dataset #{i}: nb time points: {dataset.nb_timepoints(channel)}\n"

            info_str += "\n\n"

            return info_str

    def get_metadata(self):
        """get the attributes stored in the zarr folder"""
        attrs = {}
        for dataset in self._dataset_list:
            attrs.update(dataset.get_metadata())
        return attrs

    def get_array(self, channel: str, per_z_slice: bool = False, wrap_with_dask: bool = False):
        return self._arrays[channel]

    def get_stack(self, channel: str, time_point: int, per_z_slice: bool = False, wrap_with_dask: bool = False):
        return self._arrays[channel][time_point]

    def get_projection_array(self, channel: str, axis: int, wrap_with_dask: bool = False) -> Any:
        return self._projection_arrays[f'{channel}/{axis}']

    def add_channel(self, name: str, shape: Tuple[int, ...], dtype, enable_projections: bool = True, **kwargs) -> Any:
        raise NotImplementedError("Cannot write to a joined dataset!")

    def write_stack(self, channel: str, time_point: int, stack_array: numpy.ndarray):
        raise NotImplementedError("Cannot write to a joined dataset!")

    def write_array(self, channel: str, array: numpy.ndarray):
        raise NotImplementedError("Cannot write to a joined dataset!")



from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.color.demo.demo_insert import demo_insert


def test_insert_numpy():
    with NumpyBackend():
        _test_insert()


def test_insert_cupy():
    try:
        with CupyBackend():
            _test_insert()
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def _test_insert():
    demo_insert(display=False)



from typing import Tuple, Union

from dexp.processing.backends.backend import Backend


def crop_resize_pad_color_image(image,
                                crop: Union[int, Tuple[int, ...], Tuple[Tuple[int, int], ...]] = None,
                                resize: Tuple[int, ...] = None,
                                resize_order: int = 3,
                                resize_mode: str = 'constant',
                                pad_width: Tuple[Tuple[int, int], ...] = None,
                                pad_mode: str = 'constant',
                                pad_color: Tuple[float, float, float, float] = (0, 0, 0, 0),
                                rgba_value_max: float = 255
                                ):
    """
    Crops, resizes and then pad an RGB(A) image.


    Parameters
    ----------
    image: image to resize.
    crop: Crop image by removing a given number of pixels/voxels per axis. For example: ((10,20),(10,20)) crops 10 pixels on the left for axis 0, 20 pixels from the right of axis 0, and the same for axis 2.
    resize: After cropping, the image is resized to the given shape. If any entry in the tuple is -1 then that position in the shape is automatically determined based on the existing shape to preserve aspect ratio.
    resize_order: The order of the spline interpolation, default is 3. The order has to be in the range 0-5.
    resize_mode: optional The mode parameter determines how the input array is extended beyond its boundaries. Can be: ‘reflect’, ‘constant’, ‘nearest’, ‘mirror’, ‘wrap’.
    pad_width: After cropping and resizing, padding is performed. The provided tuple is interpreted similarly to cropping.
    pad_mode: Padding mode, see numpy.pad for the available modes.
    pad_color: Padding color as tuple of normalised floats:  (R,G,B,A). Default is transparent black.
    rgba_value_max: max value for rgba values.

    Returns
    -------
    Cropped, resized, and padded image.

    """
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    # Move to backend:
    image = Backend.to_backend(image)

    # Normalisation of crop parameter:
    if crop is not None:

        if type(crop) is int:
            crop = (crop,) * (image.ndim-1)
        if type(crop[0]) is int:
            crop = tuple((c, c) for c in crop)

        # build the slice object to crop the image
        slicing = tuple(slice(l if l > 0 else None, -r if r > 0 else None) for l, r in crop)+(slice(None), )

        # Crop:
        image = image[slicing]

    # Normalise resize:
    if resize is not None:

        # computing resize factors:
        factors = tuple(ns / s for ns, s in zip(resize, image.shape[:-1]))

        # find all non negative factors:
        factors_no_negatives = tuple(factor for factor in factors if factor>0)

        # compute the average (most case all factors are equal!)
        avg_factor = sum(factors_no_negatives) / len(factors_no_negatives)

        # we replace the negative values with the average:
        factors = tuple((factor if factor > 0 else avg_factor) for factor in factors)

        # handle channel dim:
        factors = factors + (1,)

        # Resizing:
        image = sp.ndimage.zoom(input=image,
                                zoom=factors,
                                order=resize_order,
                                mode=resize_mode)

    # Number of channels:
    nb_channels = image.shape[-1]

    # Normalise pad_width:
    if pad_width is not None:

        # Adding a colored border:
        padded_channels = []
        for channel_index in range(nb_channels):
            channel = image[..., channel_index]

            value = pad_color[channel_index]*rgba_value_max
            padded_channel = xp.pad(channel,
                                    pad_width=pad_width,
                                    mode=pad_mode,
                                    constant_values=value)
            padded_channels.append(padded_channel)

        # Stacking:
        image = xp.stack(padded_channels, axis=-1)

    return image



import pytest

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.filters.kernels.gaussian import gaussian_kernel_nd


def test_gaussian_numpy():
    with NumpyBackend():
        _test_gaussian()


def test_gaussian_cupy():
    try:
        with CupyBackend():
            _test_gaussian()
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def _test_gaussian():
    kernel = gaussian_kernel_nd(ndim=3, size=17, sigma=2.0)

    assert Backend.to_numpy(kernel.sum()) == pytest.approx(1)



import json
from typing import List, Sequence, Tuple, Optional, Union

import numpy

from dexp.processing.backends.backend import Backend
from dexp.processing.registration.model.pairwise_registration_model import PairwiseRegistrationModel


class SequenceRegistrationModel:

    def __init__(self,
                 model_list: List[PairwiseRegistrationModel] = [],
                 force_numpy: bool = True
                 ):

        """ Instantiates a sequence-registration-model. A sequence registration model describes how to register jointly each n-1 D slice of a nD image.
        In effect, this object contains a sequence of registration models that describe how to transform each image in the sequence to the registered image
        that maximises registration coherence across the whole sequence.

        Parameters
        ----------
        model_list : list of registration models that
        force_numpy : when creating this object, you have the option of forcing the use of numpy array instead of the current backend arrays.
        integral : True if shifts are snapped to integer values, False otherwise

        """
        super().__init__()

        self.model_list: List[PairwiseRegistrationModel] = model_list
        if force_numpy:
            self.to_numpy()

    def append(self, model: PairwiseRegistrationModel):
        self.model_list.append(model)

    def __iadd__(self, models: Sequence[PairwiseRegistrationModel]):
        self.model_list += models

    def __len__(self):
        return self.model_list.__len__()

    def __getitem__(self, item) -> PairwiseRegistrationModel:
        return self.model_list.__getitem__(item)

    def __str__(self):
        return f"SequenceRegistrationModel(length={len(self.model_list)})"

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.model_list == other.model_list
        else:
            return False

    def to_json(self) -> str:
        models_json = list(model.to_json() for model in self.model_list)
        return json.dumps({'type': 'translation_sequence', 'model_list': models_json})

    def to_numpy(self) -> 'SequenceRegistrationModel':
        self.model_list = list(model.to_numpy() for model in self.model_list)
        return self

    def overall_confidence(self) -> float:
        confidences = list(model.overall_confidence() for model in self.model_list)
        return float(numpy.median(confidences))

    def padding(self):
        padding_list = list(model.padding() for model in self.model_list)
        overall_padding = padding_list[0]
        for padding in padding_list:
            overall_padding = tuple((max(op[0], p[0]), max(op[1], p[1])) for op, p in zip(overall_padding, padding))
        return tuple(overall_padding)

    def padded_shape(self, shape: Tuple[int, ...]):
        new_shape = tuple(s + pl + pr for s, (pl, pr) in zip(shape, self.padding()))
        return new_shape

    def apply_sequence(self,
                       image, axis: int,
                       pad_width: Optional[Union[float, Tuple[Tuple[float, float], ...]]] = None,
                       **kwargs) -> 'Array':
        """ Applies this sequence registration model to the an image sequence of given sequence axis.
            A new registered image is returned.

            Parameters
            ----------
            image: image sequence to register must be n+1 dimensional (1 dimension is the sequence axis)
            axis: sequence axis
            pad_width: Can be: (i) list of pad widths for all image edges : ((pad_left, pad_right),(...,...),..) (see numpy.pad pad_widths parameter), (ii) '0' (zero) for no padding, or (iii) None for automatic padding.
            **kwargs: parameters passthrough to the apply method of the pairwise registration model

            Returns
            -------
            stabilised image sequence

            """

        xp = Backend.get_xp_module()

        # move axis to first position, makes everything easier:
        if axis != 0:
            image = xp.moveaxis(image, axis, 0)

        if pad_width is None:
            # automatic padding
            image = xp.pad(image, pad_width=((0, 0),) + self.padding())

        elif pad_width != 0:
            # custom padding
            image = xp.pad(image, pad_width=((0, 0),) + tuple(pad_width))
        else:
            # no padding
            pass

        registered_image = xp.stack((self.apply(image[index], index=index, **kwargs) for index in range(image.shape[axis])))

        # put back axis where it belongs, if necessary:
        if axis != 0:
            registered_image = xp.moveaxis(registered_image, 0, axis)

        return registered_image

    def apply(self,
              image,
              index: int,
              pad: bool = False,
              **kwargs) -> 'Array':
        """ Applies this sequence registration model to the image at a given index.
            A new registered image is returned.


            Parameters
            ----------
            image: image to register
            index: index of image in sequence
            pad : pads the image before registration
            **kwargs : parameters passthrough to the apply method of the pairwise registration model

            Returns
            -------
            image_reg: registered image against all others in the sequence

            """
        xp = Backend.get_xp_module()
        model = self.model_list[index]

        if pad:
            padding = self.padding()
            image = xp.pad(image, pad_width=padding)
            return model.apply(image, **kwargs)

        else:
            return model.apply(image, **kwargs)

    def plot(self, path: str):
        import matplotlib.pyplot as plt

        length = len(self.model_list)
        ndim = self.model_list[0].shift_vector.shape[0]
        x = numpy.arange(0, length)
        y = numpy.zeros((length, ndim))
        for i, m in enumerate(self.model_list):
            y[i] = m.shift_vector

        fig, ax = plt.subplots()  # Create a figure and an axes.
        for axis in range(ndim):
            ax.plot(x, y[:, axis], label=f'axis {axis}')  # Plot some data on the axes.

        ax.set_xlabel('time points')  # Add an x-label to the axes.
        ax.set_ylabel('shift')  # Add a y-label to the axes.
        ax.set_title(f"{path} shifts")  # Add a title to the axes.
        ax.legend()  # Add a legend.

        plt.savefig(path + '_shifts.pdf')



import click
from arbol.arbol import aprint, asection

from dexp.cli.dexp_main import _default_workers_backend
from dexp.cli.parsing import _parse_channels, _parse_slicing, _get_output_path, _parse_devices
from dexp.datasets.open_dataset import glob_datasets
from dexp.datasets.operations.projrender import dataset_projection_rendering


@click.command()
@click.argument('input_paths', nargs=-1)
@click.option('--output_path', '-o', type=str, default=None, help='Output folder to store rendered PNGs. Default is: frames_<channel_name>')
@click.option('--channels', '-c', type=str, default=None, help='list of channels to color, all channels when ommited.')
@click.option('--slicing', '-s', type=str, default=None, help='dataset slice (TZYX), e.g. [0:5] (first five stacks) [:,0:100] (cropping in z).')
@click.option('--overwrite', '-w', is_flag=True, help='to force overwrite of target', show_default=True)  # , help='dataset slice'
@click.option('--axis', '-ax', type=int, default=0, help='Sets the projection axis: 0->Z, 1->Y, 2->X ', show_default=True)  # , help='dataset slice'
@click.option('--dir', '-di', type=int, default=-1, help='Sets the projection direction: -1 -> top to bottom, +1 -> bottom to top.', show_default=True)  # , help='dataset slice'
@click.option('--mode', '-m', type=str, default='colormax',
              help='Sets the projection mode: ‘max’: classic max projection, ‘colormax’: color max projection, i.e. color codes for depth, ‘maxcolor’ same as colormax but first does depth-coding by color and then max projects (acheives some level of transparency). ',
              show_default=True)
@click.option('--clim', '-cl', type=str, default=None, help='Sets the contrast limits, i.e. -cl 0,1000 sets the contrast limits to [0,1000]')
@click.option('--attenuation', '-at', type=float, default=0.1, help='Sets the projection attenuation coefficient, should be within [0, 1] ideally close to 0. Larger values mean more attenuation.',
              show_default=True)  # , help='dataset slice'
@click.option('--gamma', '-g', type=float, default=1.0, help='Sets the gamma coefficient pre-applied to the raw voxel values (before projection or any subsequent processing).', show_default=True)  # , help='dataset slice'
@click.option('--dlim', '-dl', type=str, default=None,
              help='Sets the depth limits. Depth limits. For example, a value of (0.1, 0.7) means that the colormap start at a normalised depth of 0.1, and ends at a normalised depth of 0.7, other values are clipped. Only used for colormax mode.',
              show_default=True)  # , help='dataset slice'
@click.option('--colormap', '-cm', type=str, default=None, help='sets colormap, e.g. viridis, gray, magma, plasma, inferno. Use a rainbow colormap such as turbo, bmy, or rainbow (recommended) for color-coded depth modes. ',
              show_default=True)
@click.option('--rgbgamma', '-cg', type=float, default=1.0, help='Gamma correction applied to the resulting RGB image. Usefull to brighten image', show_default=True)
@click.option('--transparency', '-t', is_flag=True, help='Enables transparency output when possible. Good for rendering on white (e.g. on paper).', show_default=True)
@click.option('--legendsize', '-lsi', type=float, default=1.0, help='Multiplicative factor to control size of legend. If 0, no legend is generated.', show_default=True)
@click.option('--legendscale', '-lsc', type=float, default=1.0, help='Float that gives the scale in some unit of each voxel (along the projection direction). Only in color projection modes.', show_default=True)
@click.option('--legendtitle', '-lt', type=str, default='color-coded depth (voxels)', help='Title for the color-coded depth legend.', show_default=True)
@click.option('--legendtitlecolor', '-ltc', type=str, default='1,1,1,1', help='Legend title color as a tuple of normalised floats: R, G, B, A  (values between 0 and 1).', show_default=True)
@click.option('--legendposition', '-lp', type=str, default='bottom_left', help='Position of the legend in pixels in natural order: x,y. Can also be a string: bottom_left, bottom_right, top_left, or top_right.', show_default=True)
@click.option('--legendalpha', '-la', type=float, default=1, help='Transparency for legend (1 means opaque, 0 means completely transparent)', show_default=True)
@click.option('--step', '-sp', type=int, default=1, help='Process every ‘step’ frames.', show_default=True)
@click.option('--workers', '-k', type=int, default=-1, help='Number of worker threads to spawn, if -1 then num workers = num devices', show_default=True)  #
@click.option('--workersbackend', '-wkb', type=str, default=_default_workers_backend, help='What backend to spawn workers with, can be ‘loky’ (multi-process) or ‘threading’ (multi-thread) ', show_default=True)  #
@click.option('--devices', '-d', type=str, default='0', help='Sets the CUDA devices id, e.g. 0,1,2 or ‘all’', show_default=True)  #
def projrender(input_paths,
               output_path,
               channels,
               slicing,
               overwrite,
               axis,
               dir,
               mode,
               clim,
               attenuation,
               gamma,
               dlim,
               colormap,
               rgbgamma,
               transparency,
               legendsize,
               legendscale,
               legendtitle,
               legendtitlecolor,
               legendposition,
               legendalpha,
               step,
               workers,
               workersbackend,
               devices,
               stop_at_exception=True
               ):
    """ Renders datatset using 2D projections.
    """

    input_dataset, input_paths = glob_datasets(input_paths)
    channels = _parse_channels(input_dataset, channels)
    slicing = _parse_slicing(slicing)
    devices = _parse_devices(devices)

    output_path = _get_output_path(input_paths[0], output_path, f"_{mode}_projection")

    dlim = None if dlim is None else tuple(float(strvalue) for strvalue in dlim.split(','))
    legendtitlecolor = tuple(float(v) for v in legendtitlecolor.split(','))

    if ',' in legendposition:
        legendposition = tuple(float(strvalue) for strvalue in legendposition.split(','))

    with asection(f"Projection rendering of: {input_paths} to {output_path} for channels: {channels}, slicing: {slicing} "):
        dataset_projection_rendering(input_dataset,
                                     output_path,
                                     channels,
                                     slicing,
                                     overwrite,
                                     axis,
                                     dir,
                                     mode,
                                     clim,
                                     attenuation,
                                     gamma,
                                     dlim,
                                     colormap,
                                     rgbgamma,
                                     transparency,
                                     legendsize,
                                     legendscale,
                                     legendtitle,
                                     legendtitlecolor,
                                     legendposition,
                                     legendalpha,
                                     step,
                                     workers,
                                     workersbackend,
                                     devices,
                                     stop_at_exception)

        input_dataset.close()
        aprint("Done!")
